import logging
from collections.abc import Callable
from pathlib import Path
from pydoc import locate
from typing import Literal

import typer
from ask_shell.typer_command import configure_logging
from model_lib import dump, parse_payload
from zero_3rdparty.file_utils import iter_paths

from atlas_init.cli_helper import sdk_auto_changes
from atlas_init.cli_helper.run import (
    run_binary_command_is_ok,
    run_command_exit_on_failure,
    run_command_receive_result,
)
from atlas_init.cli_helper.sdk import (
    SDK_VERSION_HELP,
    SdkVersion,
    SdkVersionUpgrade,
    find_breaking_changes,
    find_latest_sdk_version,
    format_breaking_changes,
    is_removed,
    parse_breaking_changes,
)
from atlas_init.cli_helper.tf_runner import (
    TerraformRunError,
    dump_tf_vars,
    export_outputs,
    get_tf_vars,
    run_terraform,
)
from atlas_init.repos.go_sdk import go_sdk_breaking_changes
from atlas_init.repos.path import (
    Repo,
    current_repo,
    current_repo_path,
    find_go_mod_dir,
    find_paths,
    resource_name,
)
from atlas_init.settings.config import RepoAliasNotFoundError, TestSuite
from atlas_init.settings.env_vars import (
    active_suites,
    init_settings,
)
from atlas_init.settings.env_vars_generated import AWSSettings, AtlasSettings
from atlas_init.settings.path import (
    CwdIsNoRepoPathError,
    dump_vscode_dotenv,
    repo_path_rel_path,
)
from atlas_init.typer_app import app, app_command, extra_root_commands

logger = logging.getLogger(__name__)


@app_command()
def init(context: typer.Context):
    settings = init_settings()
    extra_args = context.args
    logger.info(f"in the init command: {extra_args}")
    run_terraform(settings, "init", extra_args)


@app_command()
def plan(context: typer.Context, *, skip_outputs: bool = False):
    _plan_or_apply(context.args, "plan", skip_outputs=skip_outputs)


@app_command()
def apply(context: typer.Context, *, skip_outputs: bool = False):
    suites = _plan_or_apply(context.args, "apply", skip_outputs=skip_outputs)
    for suite in suites:
        for hook in suite.post_apply_hooks:
            logger.info(f"running post apply hook: {hook.name}")
            hook_func = locate(hook.locate)
            hook_func()  # type: ignore


def _plan_or_apply(extra_args: list[str], command: Literal["plan", "apply"], *, skip_outputs: bool) -> list[TestSuite]:  # type: ignore
    settings = init_settings(AtlasSettings, AWSSettings)
    logger.info(f"using the '{command}' command, extra args: {extra_args}")
    try:
        suites = active_suites(settings)
    except (CwdIsNoRepoPathError, RepoAliasNotFoundError) as e:
        logger.warning(repr(e))
        suites = []

    tf_vars = get_tf_vars(settings, suites)
    dump_tf_vars(settings, tf_vars)

    try:
        run_terraform(settings, command, extra_args)
    except TerraformRunError as e:
        logger.error(repr(e))  # noqa: TRY400
        raise typer.Exit(1) from e

    if not skip_outputs:
        export_outputs(settings)

    if settings.env_vars_generated.exists():
        dump_vscode_dotenv(settings.env_vars_generated, settings.env_vars_vs_code)
        logger.info(f"your .env file is ready @ {settings.env_vars_vs_code}")
    return suites


@app_command()
def destroy(context: typer.Context):
    extra_args = context.args
    settings = init_settings()
    if not settings.tf_state_path.exists():
        logger.warning(f"no terraform state found  {settings.tf_state_path}, exiting")
        return
    tf_vars = get_tf_vars(settings, [])
    dump_tf_vars(settings, tf_vars)
    try:
        run_terraform(settings, "destroy", extra_args)
    except TerraformRunError as e:
        logger.error(repr(e))  # noqa: TRY400
        return


@app_command()
def sdk_upgrade(
    old: SdkVersion = typer.Argument(help=SDK_VERSION_HELP),
    new: SdkVersion = typer.Argument(
        default_factory=find_latest_sdk_version,
        help=SDK_VERSION_HELP + "\nNo Value=Latest",
    ),
    resource: str = typer.Option("", help="for only upgrading a single resource"),
    dry_run: bool = typer.Option(False, help="only log out the changes"),
    auto_change_name: str = typer.Option("", help="any extra replacements done in the file"),
):
    SdkVersionUpgrade(old=old, new=new)
    repo_path, _ = repo_path_rel_path()
    logger.info(f"bumping from {old} -> {new} @ {repo_path}")

    sdk_breaking_changes_path = go_sdk_breaking_changes(repo_path)
    all_breaking_changes = parse_breaking_changes(sdk_breaking_changes_path, old, new)
    replacements = {
        f"go.mongodb.org/atlas-sdk/{old}/mockadmin": f"go.mongodb.org/atlas-sdk/{new}/mockadmin",
        f"go.mongodb.org/atlas-sdk/{old}/admin": f"go.mongodb.org/atlas-sdk/{new}/admin",
    }
    auto_modifier: Callable[[str, str], str] | None = None
    if auto_change_name:
        func_path = f"{sdk_auto_changes.__name__}.{auto_change_name}"
        auto_modifier = locate(func_path)  # type: ignore

    change_count = 0
    resources: set[str] = set()
    resources_breaking_changes: set[str] = set()
    for path in iter_paths(repo_path, "*.go", ".mockery.yaml"):
        text_old = path.read_text()
        if all(replace_in not in text_old for replace_in in replacements):
            continue
        r_name = resource_name(repo_path, path)
        if resource and resource != r_name:
            continue
        resources.add(r_name)
        logger.info(f"updating sdk version in {path}")
        if breaking_changes := find_breaking_changes(text_old, all_breaking_changes):
            changes_formatted = format_breaking_changes(text_old, breaking_changes)
            logger.warning(f"found breaking changes: {changes_formatted}")
            if is_removed(breaking_changes):
                resources_breaking_changes.add(r_name)
        text_new = text_old
        for replace_in, replace_out in replacements.items():
            text_new = text_new.replace(replace_in, replace_out)
        if not dry_run:
            if auto_modifier:
                text_new = auto_modifier(text_new, old)
            path.write_text(text_new)
        change_count += 1
    if change_count == 0:
        logger.warning("no changes found")
        return
    logger.info(f"changed in total: {change_count} files")
    resources_str = "\n".join(
        f"- {r} 💥" if r in resources_breaking_changes else f"- {r}" for r in sorted(resources) if r
    )
    logger.info(f"resources changed: \n{resources_str}")
    if dry_run:
        logger.warning("dry-run, no changes to go.mod")
        return
    go_mod_parent = find_go_mod_dir(repo_path)
    if not run_binary_command_is_ok("go", "mod tidy", cwd=go_mod_parent, logger=logger):
        logger.critical(f"failed to run go mod tidy in {go_mod_parent}")
        raise typer.Exit(1)


@app_command()
def pre_commit(
    skip_build: bool = typer.Option(default=False),
    skip_lint: bool = typer.Option(default=False),
    max_issues: int = typer.Option(1000, "-m", "--max"),
):
    golang_ci_lint_args = f"--max-same-issues {max_issues} --max-issues-per-linter {max_issues}"
    match current_repo():
        case Repo.CFN:
            repo_path, resource_path, r_name = find_paths()
            build_cmd = f"cd {resource_path} && make build"
            # TODO: understand why piping to grep doesn't work
            # f"golangci-lint run --path-prefix=./cfn-resources | grep {r_name}"
            format_cmd_str = (
                f"cd cfn-resources && golangci-lint run --path-prefix=./cfn-resources {golang_ci_lint_args}"
            )
        case Repo.TF:
            repo_path = current_repo_path()
            build_cmd = "make build"
            format_cmd_str = f"golangci-lint run {golang_ci_lint_args}"
        case _:
            raise NotImplementedError
    if skip_build:
        logger.warning("skipping build")
    else:
        run_command_exit_on_failure(build_cmd, cwd=repo_path, logger=logger)
    if skip_lint:
        logger.warning("skipping formatting")
    else:
        run_command_exit_on_failure(format_cmd_str, cwd=repo_path, logger=logger)


@app_command()
def repo_dump():
    code_root = Path.home() / "code"
    path_urls = {}
    for repo_git_path in iter_paths(code_root, ".git", exclude_folder_names=[".terraform"]):
        repo_path = repo_git_path.parent
        logger.info(f"repo: {repo_path}")
        url = run_command_receive_result("git remote get-url origin", cwd=repo_path, logger=logger, can_fail=True)
        if url.startswith("FAIL:"):
            continue
        path_urls[str(repo_path)] = url
    out_path = code_root / "repos.json"
    repos_json = dump(path_urls, "pretty_json")
    out_path.write_text(repos_json)


@app_command()
def repo_clone():
    repos_file = Path.home() / "code" / "repos.json"
    repo_path_json: dict[str, str] = parse_payload(repos_file)  # type: ignore
    for repo_path_str, url in repo_path_json.items():
        logger.info(f"cloning {url} @ {repo_path_str}")
        repo_path = Path(repo_path_str)
        parent_dir = repo_path.parent
        parent_dir.mkdir(parents=True, exist_ok=True)
        if repo_path.exists():
            logger.warning(f"skipping {repo_path}, already exists")
            continue
        run_command_exit_on_failure(f"git clone {url} {repo_path.name}", cwd=parent_dir, logger=logger)


def typer_main():
    extra_root_commands()
    configure_logging(app)
    app()


if __name__ == "__main__":
    typer_main()



import logging
import os
from pathlib import Path
from typing import Any

from model_lib import Entity, dump, parse_model
from mypy_boto3_cloudformation.type_defs import ParameterTypeDef
from pydantic import ConfigDict, Field
from rich import prompt
from zero_3rdparty.file_utils import clean_dir
from zero_3rdparty.dict_nested import iter_nested_key_values, update

from atlas_init.cli_cfn.files import create_sample_file, default_log_group_name
from atlas_init.cloud.aws import PascalAlias
from atlas_init.repos.cfn import CfnType, cfn_examples_dir, cfn_type_normalized

logger = logging.getLogger(__name__)
UNKNOWN_PLACEHOLDER = "UNKNOWN"


class TemplatePathNotFoundError(Exception):
    def __init__(self, type_name: str, examples_dir: Path) -> None:
        self.type_name = type_name
        self.examples_dir = examples_dir


def infer_template_path(repo_path: Path, type_name: str, stack_name: str, example_name: str = "") -> Path:
    examples_dir = cfn_examples_dir(repo_path)
    template_paths: list[Path] = []
    type_setting = f'"Type": "{type_name}"'
    for p in examples_dir.rglob("*.json"):
        if example_name and example_name != p.stem:
            continue
        if type_setting in p.read_text():
            logger.info(f"found template @ '{p.stem}': {p.parent}")
            template_paths.append(p)
    if not template_paths:
        raise TemplatePathNotFoundError(type_name, examples_dir)
    if len(template_paths) > 1:
        return choose_template_path(type_name, template_paths, stack_name)
    return template_paths[0]


def choose_template_path(type_name: str, template_paths: list[Path], stack_name: str) -> Path:
    expected_folder = cfn_type_normalized(type_name)
    if (expected_folders := [p for p in template_paths if p.parent.name == expected_folder]) and len(
        expected_folders
    ) == 1:
        logger.info(f"using template: {expected_folders[0]}")
        return expected_folders[0]
    choices = {p.stem: p for p in template_paths}
    if stack_path := choices.get(stack_name):
        logger.info(f"using template @ {stack_path} based on stack name: {stack_name}")
        return stack_path
    selected_path = prompt.Prompt("Choose example template: ", choices=list(choices))()
    return choices[selected_path]


parameters_exported_env_vars = {
    "OrgId": "MONGODB_ATLAS_ORG_ID",
    "Profile": "ATLAS_INIT_CFN_PROFILE",
    "KeyId": "MONGODB_ATLAS_ORG_API_KEY_ID",
    "TeamId": "MONGODB_ATLAS_TEAM_ID",
    "ProjectId": "MONGODB_ATLAS_PROJECT_ID",
    "AWSVpcId": "AWS_VPC_ID",
    "MongoDBAtlasProjectId": "MONGODB_ATLAS_PROJECT_ID",
    "AWSSubnetId": "AWS_SUBNET_ID",
    "AWSRegion": "AWS_REGION",
    "AppId": "MONGODB_REALM_APP_ID",
    "FunctionId": "MONGODB_REALM_FUNCTION_ID",
    "FunctionName": "MONGODB_REALM_FUNCTION_NAME",
    "ServiceId": "MONGODB_REALM_SERVICE_ID",
}

STACK_NAME_PARAM = "$STACK_NAME_PARAM$"
type_names_defaults: dict[str, dict[str, str]] = {
    "project": {
        "KeyRoles": "GROUP_OWNER",
        "TeamRoles": "GROUP_OWNER",
        STACK_NAME_PARAM: "Name",
    },
    "cluster": {
        STACK_NAME_PARAM: "ClusterName",
        "ProjectName": "Cluster-CFN-Example",
    },
    "resourcepolicy": {
        STACK_NAME_PARAM: "Name",
        "Policies": 'forbid (principal, action == cloud::Action::"project.edit",resource) when {context.project.ipAccessList.contains(ip("0.0.0.0/0"))};',
    },
    "trigger": {
        STACK_NAME_PARAM: "TriggerName",
    },
}


class CfnParameter(Entity):
    model_config = PascalAlias
    type: str
    description: str = ""
    constraint_description: str = ""
    default: str = ""
    allowed_values: list[str] = Field(default_factory=list)


class CfnResource(Entity):
    model_config = PascalAlias
    type: str
    properties: dict[str, Any] = Field(default_factory=dict)


class CfnTemplate(Entity):
    model_config = PascalAlias | ConfigDict(extra="allow")
    parameters: dict[str, CfnParameter]
    resources: dict[str, CfnResource]

    @classmethod
    def read_template_types(cls, template_path: Path, prefix: str = CfnType.MONGODB_ATLAS_CFN_TYPE_PREFIX) -> set[str]:
        cfn_template = parse_model(template_path, t=CfnTemplate)
        return {r.type for r in cfn_template.resources.values() if r.type.startswith(prefix)}

    def find_resource(self, type_name: str) -> CfnResource:
        for r in self.resources.values():
            if r.type == type_name:
                return r
        raise ValueError(f"resource not found: {type_name}")

    def normalized_type_name(self, type_name: str) -> str:
        assert self.find_resource(type_name)
        return cfn_type_normalized(type_name)

    def get_resource_properties(self, type_name: str, parameters: list[ParameterTypeDef]) -> dict:
        resource = self.find_resource(type_name)
        properties = resource.properties
        for param in parameters:
            key = param.get("ParameterKey")
            assert key
            if key not in properties:
                key_found = next(
                    (
                        maybe_key
                        for maybe_key, value in properties.items()
                        if isinstance(value, dict) and value.get("Ref") == key
                    ),
                    None,
                )
                err_msg = f"unable to find parameter {key} in resource {type_name}, can happen if there are template parameters not used for {type_name}"
                if key_found is None:
                    logger.warning(err_msg)
                    continue
                key = key_found
            param_value = param.get("ParameterValue", "")
            assert param_value
            properties[key] = param_value
        return properties


class CfnTemplateUnknownParametersError(Exception):
    def __init__(self, unknown_params: list[str]) -> None:
        self.unknown_params = unknown_params


def infer_template_parameters(
    path: Path, type_name: str, stack_name: str, explicit_params: dict[str, Any]
) -> list[ParameterTypeDef]:
    cfn_template = parse_model(path, t=CfnTemplate)
    parameters_dict: dict[str, Any] = {key: UNKNOWN_PLACEHOLDER for key in cfn_template.parameters.keys()}
    type_defaults = type_names_defaults.get(cfn_template.normalized_type_name(type_name), {})
    if stack_name_param := type_defaults.pop(STACK_NAME_PARAM, None):
        type_defaults[stack_name_param] = stack_name
    for param_name, param in cfn_template.parameters.items():
        explicit_value = explicit_params.get(param_name)
        if explicit_value is not None:
            logger.info(f"using explicit value for {param_name}={explicit_value}")
            parameters_dict[param_name] = explicit_value
            continue
        if type_default := type_defaults.get(param_name):
            logger.info(f"using type default for {param_name}={type_default}")
            parameters_dict[param_name] = type_default
            continue
        if env_key := parameters_exported_env_vars.get(param_name):
            if env_value := os.environ.get(env_key):
                logger.info(f"using {env_key} to fill parameter: {param_name}")
                parameters_dict[param_name] = env_value
                continue
        if set(param.allowed_values) == {"true", "false"}:
            logger.info(f"using default false for {param_name}")
            parameters_dict[param_name] = "false"
            continue
        if default := param.default:
            logger.info(f"using default for {param_name}={default}")
            parameters_dict[param_name] = default
            continue
    if unknown_params := {key for key, value in parameters_dict.items() if value == UNKNOWN_PLACEHOLDER}:
        raise CfnTemplateUnknownParametersError(sorted(unknown_params))
    return [{"ParameterKey": key, "ParameterValue": value} for key, value in parameters_dict.items()]


def dump_resource_to_file(
    inputs_dir: Path,
    template_path: Path,
    type_name: str,
    parameters: list[ParameterTypeDef],
) -> Path:
    cfn_template = parse_model(template_path, t=CfnTemplate)
    properties = cfn_template.get_resource_properties(type_name, parameters)
    clean_dir(inputs_dir, recreate=True)
    dest_path = inputs_dir / "inputs_1_create.json"
    dest_json = dump(properties, "pretty_json")
    dest_path.write_text(dest_json)
    return dest_path


def dump_sample_file(
    samples_dir: Path,
    template_path: Path,
    type_name: str,
    parameters: list[ParameterTypeDef],
):
    cfn_template = parse_model(template_path, t=CfnTemplate)
    samples_path = samples_dir / template_path.stem / "create.json"
    create_sample_file(
        samples_path,
        default_log_group_name(CfnType.resource_name(type_name)),
        cfn_template.get_resource_properties(type_name, parameters),
        prev_resource_state={},
    )
    return samples_path


def modify_resource_with_params(resource: CfnResource, resource_params: dict[str, Any]) -> None:
    updates: dict[str, tuple[str, Any]] = {}
    resource_properties = resource.properties
    for path, value in iter_nested_key_values(resource_properties, include_list_indexes=True):
        if not isinstance(value, dict):
            continue
        if "Ref" not in value:
            continue
        param_name = value["Ref"]
        assert isinstance(param_name, str), f"Ref must be a string, {path}, got={param_name!r}"
        if param_value := resource_params.get(param_name):
            updates[param_name] = (path, param_value)
        else:
            logger.warning(f"unable to find parameter {param_name} in resource params, path={path}")

    for param_name, param_value in resource_params.items():
        if update_path_value := updates.get(param_name):
            update(resource_properties, *update_path_value)
        else:
            logger.warning(f"No ref found for {param_name} assumming top level on resource")
            resource_properties[param_name] = param_value



from queue import Queue
from atlas_init.cli_tf.codegen.models import ResourceConfig
from atlas_init.cli_tf.openapi import OpenapiSchema
from atlas_init.cli_tf.schema_v2 import SchemaResource


def minimal_api_spec_simplified(resource: ResourceConfig, full_spec: OpenapiSchema) -> OpenapiSchema:
    minimal_spec = OpenapiSchema(
        openapi=full_spec.openapi,
        info={"description": "minimal spec", "version": full_spec.info["version"], "title": full_spec.info["title"]},
        paths={},
        components={"schemas": {}, "parameters": {}},
    )
    include_refs: Queue[str] = Queue()
    seen_refs: set[str] = set()

    for path in resource.paths:
        path_dict = minimal_spec.paths[path] = full_spec.paths[path]
        remove_non_2xx_responses(path_dict)
        for ref in full_spec.method_refs(path):
            minimal_spec.add_schema_ref(ref, full_spec.resolve_ref(ref))
            include_refs.put(ref)
        for ref in full_spec.parameter_refs(path):
            minimal_spec.add_schema_ref(ref, full_spec.resolve_ref(ref))
            include_refs.put(ref)

    def add_from_resource_ref(ref_resource: SchemaResource) -> None:
        for attribute in ref_resource.attributes.values():
            if attribute.schema_ref:
                minimal_spec.add_schema_ref(attribute.schema_ref, full_spec.resolve_ref(attribute.schema_ref))
                include_refs.put(attribute.schema_ref)
            if attribute.parameter_ref:
                minimal_spec.add_schema_ref(
                    attribute.parameter_ref,
                    full_spec.resolve_ref(attribute.parameter_ref),
                )
            if ref := attribute.additional_properties_ref:
                minimal_spec.add_schema_ref(ref, full_spec.resolve_ref(ref))
                include_refs.put(ref)
        for ref in ref_resource.extra_refs():
            minimal_spec.add_schema_ref(ref, full_spec.resolve_ref(ref))
            include_refs.put(ref)

    while not include_refs.empty():
        ref = include_refs.get()
        if ref in seen_refs:
            continue
        seen_refs.add(ref)
        if ref.startswith(full_spec.SCHEMAS_PREFIX):
            ref_resource = full_spec.schema_ref_component(ref, set())
            add_from_resource_ref(ref_resource)
        else:
            param_name = ref.split("/")[-1]
            minimal_spec.components["parameters"][param_name] = full_spec.resolve_ref(ref)
    sorted_components = sorted(minimal_spec.components["schemas"].items())
    sorted_parameters = sorted(minimal_spec.components["parameters"].items())
    modify_schema_properties(sorted_components)
    minimal_spec.components["schemas"] = dict(sorted_components)
    minimal_spec.components["parameters"] = dict(sorted_parameters)
    return minimal_spec


def modify_schema_properties(schema_properties: list[tuple[str, dict]]):
    for _, schema_values in schema_properties:
        properties = schema_values.get("properties", {})
        for _, prop_values in properties.items():
            prop_values.pop("name", None)  # Remove 'name' field if it exists


def remove_non_2xx_responses(path_dict: dict) -> None:
    for method_dict in path_dict.values():
        method_dict["responses"] = {
            code: response for code, response in method_dict.get("responses", {}).items() if code.startswith("2")
        }



from __future__ import annotations

import logging
import re
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Literal, Self

from atlas_init.cli_tf.hcl.parser import (
    Block,
    ResourceBlock,
    hcl_attrs,
    iter_blocks,
    iter_resource_blocks,
)

logger = logging.getLogger(__name__)
INDENT = "  "
LEGACY_CLUSTER_TYPE = "mongodbatlas_cluster"
NEW_CLUSTER_TYPE = "mongodbatlas_advanced_cluster"


def indent(level: int, line: str) -> str:
    return INDENT * level + line


def convert_clusters(tf_dir: Path, out_path: Path | None = None) -> dict[tuple[Path, ResourceBlock], str]:
    new_filepath = out_path or tf_dir / "conversion_cluster_adv_cluster.tf"
    new_content: list[str] = []
    all_blocks = {}
    for path in tf_dir.glob("*.tf"):
        legacy = path.read_text()
        logger.info(f"checking for cluster blocks in {path}")
        new_blocks = parse_and_convert_cluster_blocks(legacy)
        if not new_blocks:
            continue
        all_blocks |= {(path, block): new_config for block, new_config in new_blocks.items()}
        new_content.append(f"# file @ {path}")
        for cluster, new_config in new_blocks.items():
            logger.info(f"found {cluster} to migrate in {path}")
            new_content.extend((f"# {cluster}", new_config))
    assert new_content, "unable to find any cluster resources"
    new_filepath.write_text("\n".join(new_content))
    return all_blocks


def convert_cluster_config(hcl_config: str) -> str:
    """
    For the given HCL config, convert the `mongodbatlas_cluster` resource to `mongodbatlas_advanced_cluster`.

    Strategy:
    - support find all resource blocks, e.g. `resource "mongodbatlas_cluster" "project_cluster_free"`
    - for each resource block
        - rename the resource_type to `mongodbatlas_advanced_cluster`
        - Iterate through all root attributes and copy them to the new resource.
        - Iterate through all the block attributes
            - If the block attribute is `replication_specs`:
                - Iterate through all the attributes and copy them.
                - Iterate through all the nested_blocks and copy them
            - Otherwise, copy the block attribute as is.
        for every copy, lookup:
            - if the key has been renamed in the new resource.
    Special attributes:
    - disk_size_gb (only set on root in legacy but in electable_specs in new)
    - provider_name (only set on root in legacy but in replication_specs in new)
    - node_counts are specs in new
    - auto_scaling_xx has moved to a block
    """
    converted_blocks = parse_and_convert_cluster_blocks(hcl_config)
    logger.info(f"found {len(converted_blocks)} blocks to replace")
    for block, new_block in converted_blocks.items():
        hcl_config = hcl_config.replace(block.hcl, new_block)
    return hcl_config


def parse_and_convert_cluster_blocks(hcl_config: str) -> dict[ResourceBlock, str]:
    return {
        block: convert_cluster_block(block)
        for block in iter_resource_blocks(hcl_config)
        if block.type == LEGACY_CLUSTER_TYPE
    }


_removed_attributes_root = {
    "provider_name",
    "provider_instance_size_name",
    "auto_scaling_disk_gb_enabled",
    "auto_scaling_compute_enabled",
    "provider_auto_scaling_compute_min_instance_size",
    "provider_auto_scaling_compute_max_instance_size",
    "auto_scaling_compute_scale_down_enabled",
    "backing_provider_name",
    "provider_disk_iops",
    "provider_encrypt_ebs_volume",
    "provider_volume_type",
    "provider_region_name",
    "replication_factor",
    "num_shards",
}
_removed_attributes_region_config = {
    "electable_nodes",
    "read_only_nodes",
    "analytics_nodes",
}

_renamed_attributes = {
    "cloud_backup": "backup_enabled",
}


def attribute_migration(
    block_name: Literal["root", "", "region_config"], key: str, value: str
) -> tuple[str, str] | None:
    if block_name == "root":
        if key in _removed_attributes_root:
            return None
        key = _renamed_attributes.get(key, key)
        return key, value
    if block_name == "region_config":
        if key in _removed_attributes_region_config:
            return None
        return key, value
    return key, value


def write_attributes(
    level: int,
    attributes: dict[str, str],
    block_name: Literal["root", "", "region_config"] = "",
) -> list[str]:
    lines = []
    for key, value in attributes.items():
        migrated_key_value = attribute_migration(block_name, key, value)
        if not migrated_key_value:
            continue
        new_key, new_value = migrated_key_value
        lines.append(f"{'  ' * level}{new_key} = {new_value}")
    return lines


@dataclass
class ClusterMigContext:
    # root level
    provider_name: str = ""
    provider_instance_size_name: str = ""
    auto_scaling_disk_gb_enabled: str = ""
    auto_scaling_compute_enabled: str = ""
    provider_auto_scaling_compute_min_instance_size: str = ""
    provider_auto_scaling_compute_max_instance_size: str = ""
    auto_scaling_compute_scale_down_enabled: str = ""
    backing_provider_name: str = ""
    provider_disk_iops: str = ""
    provider_encrypt_ebs_volume: str = ""
    provider_volume_type: str = ""
    provider_region_name: str = ""

    # region_config
    electable_nodes: str = ""
    read_only_nodes: str = ""
    analytics_nodes: str = ""

    def add_region_config(self, region_config: dict[str, str]) -> Self:
        kwargs = asdict(self) | as_mig_context_kwargs(region_config)
        return type(self)(**kwargs)

    @property
    def auto_scaling_lines(self) -> list[str]:
        auto_scaling_block = {}
        if self.auto_scaling_disk_gb_enabled:
            auto_scaling_block["disk_gb_enabled"] = self.auto_scaling_disk_gb_enabled
        if self.auto_scaling_compute_enabled:
            auto_scaling_block["compute_enabled"] = self.auto_scaling_compute_enabled
        if self.provider_auto_scaling_compute_min_instance_size:
            auto_scaling_block["compute_min_instance_size"] = self.provider_auto_scaling_compute_min_instance_size
        if self.provider_auto_scaling_compute_max_instance_size:
            auto_scaling_block["compute_max_instance_size"] = self.provider_auto_scaling_compute_max_instance_size
        if self.auto_scaling_compute_scale_down_enabled:
            auto_scaling_block["compute_scale_down_enabled"] = self.auto_scaling_compute_scale_down_enabled
        if not auto_scaling_block:
            return []
        return [
            indent(3, "auto_scaling {"),
            *write_attributes(4, auto_scaling_block),
            indent(3, "}"),
        ]

    def hardware_spec(self, node_count: str) -> dict[str, str]:
        hardware_spec = {}
        if node_count:
            hardware_spec["node_count"] = node_count
        if self.provider_instance_size_name:
            hardware_spec["instance_size"] = self.provider_instance_size_name
        if self.provider_disk_iops:
            hardware_spec["disk_iops"] = self.provider_disk_iops
        if self.provider_volume_type:
            hardware_spec["ebs_volume_type"] = self.provider_volume_type
        elif "AWS" in self.provider_name:
            hardware_spec["ebs_volume_type"] = '"STANDARD"'
        return hardware_spec

    @property
    def electable_spec_lines(self) -> list[str]:
        if not self.electable_nodes and not self.provider_instance_size_name:
            return []

        electable_block = self.hardware_spec(self.electable_nodes)
        return [
            indent(3, "electable_specs {"),
            *write_attributes(4, electable_block),
            indent(3, "}"),
        ]

    @property
    def analytics_spec_lines(self) -> list[str]:
        if not self.analytics_nodes:
            return []
        analytics_block = self.hardware_spec(self.analytics_nodes)
        return [
            indent(3, "analytics_specs {"),
            *write_attributes(4, analytics_block),
            indent(3, "}"),
        ]

    @property
    def read_only_spec_lines(self) -> list[str]:
        if not self.read_only_nodes:
            return []
        read_only_block = self.hardware_spec(self.read_only_nodes)
        return [
            indent(3, "read_only_specs {"),
            *write_attributes(4, read_only_block),
            indent(3, "}"),
        ]

    def region_config_lines(self, attributes: dict[str, str]) -> list[str]:
        if self.provider_region_name:
            attributes.setdefault("region_name", self.provider_region_name)
        if self.provider_name:
            attributes.setdefault("provider_name", self.provider_name)
        if self.backing_provider_name:
            attributes.setdefault("backing_provider_name", self.backing_provider_name)
        return [
            indent(2, "region_configs {"),
            *write_attributes(3, attributes, "region_config"),
            *self.auto_scaling_lines,
            *self.electable_spec_lines,
            *self.analytics_spec_lines,
            *self.read_only_spec_lines,
            indent(2, "}"),
        ]


_mig_context_fields = set(asdict(ClusterMigContext()))


def as_mig_context_kwargs(attributes: dict[str, str]) -> dict[str, str]:
    return {k: v for k, v in attributes.items() if k in _mig_context_fields}


_default_replication_spec_legacy = """\
  replication_specs {
    regions_config {
      priority = 7
    }
  }"""


def default_replication_spec(line_start: int) -> Block:
    hcl = _default_replication_spec_legacy
    return Block(
        name="replication_specs",
        level=1,
        line_start=line_start,
        line_end=line_start + len(hcl.split("\n")),
        hcl=hcl,
    )


_dynamic_pattern = re.compile(r"dynamic\s+\"[^\"]+\"\s+{")


def convert_cluster_block(root_block: ResourceBlock) -> str:
    if _dynamic_pattern.search(root_block.hcl):
        err_msg = f"dynamic block found for {root_block}, currently dynamic blocks are not supported"
        raise ValueError(err_msg)
    root_blocks = list(iter_blocks(root_block))
    attributes_root = hcl_attrs(root_block)
    attributes_root.setdefault("cluster_type", '"REPLICASET"')
    cluster_content = [
        f'resource "{NEW_CLUSTER_TYPE}" "{root_block.name}" {{',
    ]
    cluster_content.extend(write_attributes(1, attributes_root, "root"))
    mig_context = ClusterMigContext(**as_mig_context_kwargs(attributes_root))
    replication_spec_blocks = get_replication_specs(root_block)
    if not replication_spec_blocks:
        line_start = len(root_block.content_lines())
        root_blocks.append(default_replication_spec(line_start))
    for block in root_blocks:
        if block.name == "replication_specs":
            cluster_content.extend(write_replication_spec(block, mig_context))
        elif block.name == "lifecycle":
            cluster_content.extend(write_lifecycle(block))
        else:
            cluster_content.append(block.hcl)
    cluster_content.append("}")
    return "\n".join(cluster_content)


def write_replication_spec(block: Block, mig_context: ClusterMigContext) -> list[str]:
    nested_blocks = list(iter_blocks(block))
    attributes = hcl_attrs(block)
    lines = [
        "  replication_specs {",
        *write_attributes(2, attributes),
    ]
    for block in nested_blocks:
        if block.name == "regions_config":
            lines.extend(write_region_config(block, mig_context))
        else:
            lines.append(block.hcl)
    lines.append("  }")
    return lines


def write_region_config(block: Block, mig_context: ClusterMigContext) -> list[str]:
    attributes = hcl_attrs(block)
    region_config_mig = mig_context.add_region_config(attributes)
    return region_config_mig.region_config_lines(attributes)


def get_replication_specs(resource: Block) -> list[Block]:
    return [block for block in iter_blocks(resource) if block.name == "replication_specs"]


def write_lifecycle(lifecycle: Block) -> list[str]:
    attributes = hcl_attrs(lifecycle)
    if ignore_changes := attributes.get("ignore_changes", ""):
        for ignored_name in _removed_attributes_root:
            regex = re.compile(f"{ignored_name},?")
            ignore_changes = regex.sub("", ignore_changes)
        attributes["ignore_changes"] = ignore_changes
    blocks = list(iter_blocks(lifecycle))
    return [
        indent(1, "lifecycle {"),
        *write_attributes(2, attributes),
        *(block.hcl for block in blocks),
        indent(1, "}"),
    ]



import os
import subprocess  # nosec
import sys
from logging import Logger
from pathlib import Path
from shutil import which
from tempfile import TemporaryDirectory
from typing import IO

import typer
from zero_3rdparty.id_creator import simple_id

LOG_CMD_PREFIX = "running: '"


def run_command_is_ok(
    cmd: str,
    env: dict | None,
    cwd: Path | str,
    logger: Logger,
    output: IO | None = None,
    *,
    dry_run: bool = False,
) -> bool:
    env = env or {**os.environ}
    logger.info(f"{LOG_CMD_PREFIX}{cmd}' from '{cwd}'")
    if dry_run:
        return True
    output = output or sys.stdout  # type: ignore
    exit_code = subprocess.call(
        cmd,
        stdin=sys.stdin,
        stderr=sys.stderr,
        stdout=output,
        cwd=cwd,
        env=env,
        shell=True,  # noqa: S602 # We control the calls to this function and don't suspect any shell injection #nosec
    )
    is_ok = exit_code == 0
    if is_ok:
        logger.info(f"success 🥳 '{cmd}'\n")  # adds extra space to separate runs
    else:
        logger.error(f"error 💥, exit code={exit_code}, '{cmd}'")
    return is_ok


def run_binary_command_is_ok(
    binary_name: str, command: str, cwd: Path, logger: Logger, env: dict | None = None, *, dry_run: bool = False
) -> bool:
    env = env or {**os.environ}
    bin_path = find_binary_on_path(binary_name, logger, allow_missing=dry_run) or binary_name
    return run_command_is_ok(
        f"{bin_path} {command}",
        env=env,
        cwd=cwd,
        logger=logger,
        dry_run=dry_run,
    )


def find_binary_on_path(binary_name: str, logger: Logger, *, allow_missing: bool = False) -> str:
    if bin_path := which(binary_name):
        return bin_path
    if allow_missing:
        logger.warning(f"binary '{binary_name}' not found on $PATH")
        return ""
    logger.critical(f"please install '{binary_name}'")
    raise typer.Exit(1)


def run_command_exit_on_failure(
    cmd: str, cwd: Path | str, logger: Logger, env: dict | None = None, *, dry_run: bool = False
) -> None:
    if not run_command_is_ok(cmd, cwd=cwd, env=env, logger=logger, dry_run=dry_run):
        logger.critical("command failed, see output 👆")
        raise typer.Exit(1)


def run_command_receive_result(
    command: str, cwd: Path, logger: Logger, env: dict | None = None, *, can_fail: bool = False
) -> str:
    with TemporaryDirectory() as temp_dir:
        result_file = Path(temp_dir) / "file"
        with open(result_file, "w") as file:
            is_ok = run_command_is_ok(command, env=env, cwd=cwd, logger=logger, output=file)
        output_text = result_file.read_text().strip()
    if not is_ok:
        if can_fail:
            logger.warning(f"command failed {command}, {output_text}")
            return f"FAIL: {output_text}"
        logger.critical(f"command failed {command}, {output_text}")
        raise typer.Exit(1)
    return output_text


def run_command_is_ok_output(command: str, cwd: Path, logger: Logger, env: dict | None = None) -> tuple[bool, str]:
    with TemporaryDirectory() as temp_dir:
        result_file = Path(temp_dir) / f"{simple_id()}.txt"
        with open(result_file, "w") as file:
            is_ok = run_command_is_ok(command, env=env, cwd=cwd, logger=logger, output=file)
        output_text = result_file.read_text().strip()
    return is_ok, output_text


def add_to_clipboard(clipboard_content: str, logger: Logger):
    if pb_binary := find_binary_on_path("pbcopy", logger, allow_missing=True):
        subprocess.run(pb_binary, text=True, input=clipboard_content, check=True)  # nosec
    else:
        logger.warning("pbcopy not found on $PATH")



from collections import defaultdict
import logging
from contextlib import suppress
from pathlib import Path
from typing import NamedTuple
from lark import Token, Transformer, Tree, UnexpectedToken, v_args
from hcl2.transformer import Attribute, DictTransformer
from hcl2.api import reverse_transform, writes, parses
import rich

logger = logging.getLogger(__name__)


def update_attribute_object_str_value_for_block(
    tree: Tree, block_name: str, block_transformer: DictTransformer
) -> Tree:
    class BlockUpdater(Transformer):
        @v_args(tree=True)
        def block(self, block_tree: Tree) -> Tree:
            current_block_name = _identifier_name(block_tree)
            if current_block_name == block_name:
                tree_dict = block_transformer.transform(tree)
                tree_modified = reverse_transform(tree_dict)
                assert isinstance(tree_modified, Tree)
                body_tree = tree_modified.children[0]
                assert isinstance(body_tree, Tree)
                block_tree = body_tree.children[0]
                assert isinstance(block_tree, Tree)
                return block_tree
            return block_tree

    return BlockUpdater().transform(tree)


class AttributeChange(NamedTuple):
    attribute_name: str
    old_value: str | None
    new_value: str


def attribute_transfomer(attr_name: str, obj_key: str, new_value: str) -> tuple[DictTransformer, list[AttributeChange]]:
    changes: list[AttributeChange] = []

    class AttributeTransformer(DictTransformer):
        def attribute(self, args: list) -> Attribute:
            found_attribute = super().attribute(args)
            if found_attribute.key == attr_name:
                attribute_value = found_attribute.value
                if not isinstance(attribute_value, dict):
                    raise ValueError(f"Expected a dict for attribute {attr_name}, but got {type(attribute_value)}")
                old_value = attribute_value.get(obj_key)
                if old_value == new_value:
                    return found_attribute
                changes.append(AttributeChange(attr_name, old_value, new_value))
                return Attribute(attr_name, found_attribute.value | {obj_key: new_value})
            return found_attribute

    return AttributeTransformer(with_meta=True), changes


def variable_reader(tree: Tree) -> dict[str, str | None]:
    """
    Reads the variable names from a parsed HCL2 tree.
    Returns a variable_name -> description, None if no description is found.
    """
    variables: dict[str, str | None] = {}

    class DescriptionReader(DictTransformer):
        def __init__(self, with_meta: bool = False, *, name: str):
            super().__init__(with_meta)
            self.name = name
            self.description: str | None = None

        def attribute(self, args: list) -> Attribute:
            name = args[0]
            if name == "description":
                description = _parse_attribute_value(args)
                self.description = description
            return super().attribute(args)

    class BlockReader(Transformer):
        @v_args(tree=True)
        def block(self, block_tree: Tree) -> Tree:
            current_block_name = _identifier_name(block_tree)
            if current_block_name == "variable":
                variable_name = token_name(block_tree.children[1])
                reader = DescriptionReader(name=variable_name)
                reader.transform(block_tree)
                variables[variable_name] = reader.description
            return block_tree

    BlockReader().transform(tree)
    return variables


def _parse_attribute_value(args: list) -> str:
    description = args[-1]
    return token_name(description) if isinstance(description, Token) else description.strip('"')


def resource_types_vars_usage(tree: Tree) -> dict[str, dict[str, str]]:
    """
    Reads the resource types and their variable usages from a parsed HCL2 tree.
    Returns a dictionary where keys are resource type names and values are dictionaries
    of variable names and the attribute paths they are used in.
    """
    resource_types: dict[str, dict[str, str]] = defaultdict(dict)

    class ResourceBlockAttributeReader(DictTransformer):
        def __init__(self, with_meta: bool = False, resource_type: str = ""):
            self.resource_type = resource_type
            resource_types.setdefault(self.resource_type, {})
            super().__init__(with_meta)

        def attribute(self, args: list) -> Attribute:
            try:
                value = _parse_attribute_value(args)
            except AttributeError:
                return super().attribute(args)
            if value.startswith("var."):
                variable_name = value[4:]
                resource_types[self.resource_type][variable_name] = args[0]
            return super().attribute(args)

    class BlockReader(Transformer):
        @v_args(tree=True)
        def block(self, block_tree: Tree) -> Tree:
            block_resource_name = _block_resource_name(block_tree)
            if block_resource_name is not None:
                ResourceBlockAttributeReader(with_meta=True, resource_type=block_resource_name).transform(block_tree)
            return block_tree

    BlockReader().transform(tree)
    return resource_types


def variable_usages(variable_names: set[str], tree: Tree) -> dict[str, set[str]]:
    usages = defaultdict(set)
    current_resource_type = None

    class ResourceBlockAttributeReader(DictTransformer):
        def attribute(self, args: list) -> Attribute:
            attr_value = args[-1]
            if isinstance(attr_value, str) and attr_value.startswith("var."):
                variable_name = attr_value[4:]
                if variable_name in variable_names:
                    assert current_resource_type is not None, "current_resource_type should not be None"
                    usages[variable_name].add(current_resource_type)
            return super().attribute(args)

    class BlockReader(Transformer):
        @v_args(tree=True)
        def block(self, block_tree: Tree) -> Tree:
            block_resource_name = _block_resource_name(block_tree)
            if block_resource_name is not None and block_resource_name.startswith("mongodbatlas_"):
                nonlocal current_resource_type
                current_resource_type = block_resource_name
                ResourceBlockAttributeReader().transform(block_tree)
            return block_tree

    BlockReader().transform(tree)
    return usages


def _identifier_name(tree: Tree) -> str | None:
    with suppress(Exception):
        identifier_tree = tree.children[0]
        assert identifier_tree.data == "identifier"
        name_token = identifier_tree.children[0]
        assert isinstance(name_token, Token)
        if name_token.type == "NAME":
            return name_token.value


def _block_resource_name(tree: Tree) -> str | None:
    block_name = _identifier_name(tree)
    if block_name != "resource":
        return None
    token = tree.children[1]
    return token_name(token)


def token_name(token):
    assert isinstance(token, Token)
    token_value = token.value
    assert isinstance(token_value, str)
    return token_value.strip('"')


def write_tree(tree: Tree) -> str:
    return writes(tree)


def print_tree(path: Path) -> None:
    tree = safe_parse(path)
    if tree is None:
        return
    logger.info("=" * 10 + f"tree START of {path.parent.name}/{path.name}" + "=" * 10)
    rich.print(tree)
    logger.info("=" * 10 + f"tree END of {path.parent.name}/{path.name}" + "=" * 10)


def safe_parse(path: Path) -> Tree | None:
    try:
        return parses(path.read_text())  # type: ignore
    except UnexpectedToken as e:
        logger.warning(f"failed to parse {path}: {e}")



import logging
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Callable

import hcl2
from lark import Token, Tree

from atlas_init.cli_tf.hcl.modifier2 import safe_parse

logger = logging.getLogger(__name__)

BLOCK_TYPE_VARIABLE = "variable"
BLOCK_TYPE_OUTPUT = "output"


def process_token(node: Token, indent=0):
    debug_log(f"token:{node.type}:{node.value}", indent)
    return deepcopy(node)


def debug_log(message: str, depth=0):
    logger.debug("  " * depth + message.rstrip("\n"))


def is_identifier_block_type(tree: Tree | Token, block_type: str) -> bool:
    if not isinstance(tree, Tree):
        return False
    try:
        return tree.children[0].value == block_type  # type: ignore
    except (IndexError, AttributeError):
        return False


def is_block_type(tree: Tree, block_type: str) -> bool:
    try:
        return tree.data == "block" and is_identifier_block_type(tree.children[0], block_type)
    except (IndexError, AttributeError):
        return False


def update_description(tree: Tree, new_descriptions: dict[str, str], existing_names: dict[str, list[str]]) -> Tree:
    new_children = tree.children.copy()
    variable_body = new_children[2]
    assert variable_body.data == "body"
    name = token_name(new_children[1])
    old_description = read_description_attribute(variable_body)
    existing_names[name].append(old_description)
    new_description = new_descriptions.get(name, "")
    if not new_description:
        debug_log(f"no description found for variable {name}", 0)
        return tree
    new_children[2] = update_body_with_description(variable_body, new_description)
    return Tree(tree.data, new_children)


def token_name(token: Token | Tree) -> str:
    if isinstance(token, Token):
        return token.value.strip('"')
    if isinstance(token, Tree) and token.data == "identifier":
        return token.children[0].value.strip('"')  # type: ignore
    err_msg = f"unexpected token type {type(token)} for token name"
    raise ValueError(err_msg)


def has_attribute_description(maybe_attribute: Token | Tree) -> bool:
    if not isinstance(maybe_attribute, Tree):
        return False
    return maybe_attribute.data == "attribute" and maybe_attribute.children[0].children[0].value == "description"  # type: ignore


def update_body_with_description(tree: Tree, new_description: str) -> Tree:
    new_description = new_description.replace('"', '\\"')
    new_children = tree.children.copy()
    found_description = False
    for i, maybe_attribute in enumerate(new_children):
        if has_attribute_description(maybe_attribute):
            found_description = True
            new_children[i] = create_description_attribute(new_description)
    if not found_description:
        new_children.insert(0, new_line())
        new_children.insert(1, create_description_attribute(new_description))
    return Tree(tree.data, new_children)


def new_line() -> Tree:
    return Tree(
        Token("RULE", "new_line_or_comment"),
        [Token("NL_OR_COMMENT", "\n  ")],
    )


def read_description_attribute(tree: Tree) -> str:
    return next(
        (
            token_name(maybe_attribute.children[-1].children[0])
            for maybe_attribute in tree.children
            if has_attribute_description(maybe_attribute)
        ),
        "",
    )


def create_description_attribute(description_value: str) -> Tree:
    children = [
        Tree(Token("RULE", "identifier"), [Token("NAME", "description")]),
        Token("EQ", " ="),
        Tree(Token("RULE", "expr_term"), [Token("STRING_LIT", f'"{description_value}"')]),
    ]
    return Tree(Token("RULE", "attribute"), children)


def process_generic(
    node: Tree,
    tree_match: Callable[[Tree], bool],
    tree_call: Callable[[Tree], Tree],
    depth=0,
):
    new_children = []
    debug_log(f"tree:{node.data}", depth)
    for child in node.children:
        if isinstance(child, Tree):
            if tree_match(child):
                child = tree_call(child)
            new_children.append(process_generic(child, tree_match, tree_call, depth + 1))
        else:
            new_children.append(process_token(child, depth + 1))
    return Tree(node.data, new_children)


def process_descriptions(
    node: Tree,
    name_updates: dict[str, str],
    existing_names: dict[str, list[str]],
    depth=0,
    *,
    block_type: str,
) -> Tree:
    def tree_match(tree: Tree) -> bool:
        return is_block_type(tree, block_type)

    def tree_call(tree: Tree) -> Tree:
        return update_description(tree, name_updates, existing_names)

    return process_generic(
        node,
        tree_match,
        tree_call,
        depth=depth,
    )


def update_descriptions(tf_path: Path, new_names: dict[str, str], block_type: str) -> tuple[str, dict[str, list[str]]]:
    tree = safe_parse(tf_path)
    if tree is None:
        return "", {}
    existing_descriptions = defaultdict(list)
    new_tree = process_descriptions(
        tree,
        new_names,
        existing_descriptions,
        block_type=block_type,
    )
    new_tf = hcl2.writes(new_tree)  # type: ignore
    return new_tf, existing_descriptions


def _block_name_body(tree: Tree) -> tuple[str, Tree]:
    try:
        _, name_token, body = tree.children
        name = token_name(name_token)
    except (IndexError, AttributeError) as e:
        raise ValueError("unexpected block structure") from e
    return name, body


def _read_attribute(tree_body: Tree, attribute_name: str) -> Tree | None:
    for attribute in tree_body.children:
        if not isinstance(attribute, Tree):
            continue
        if attribute.data != "attribute":
            continue
        attr_identifier, _, attr_value = attribute.children
        if token_name(attr_identifier.children[0]) != attribute_name:
            continue
        return attr_value


def _is_object(tree_body: Tree) -> bool:
    if not isinstance(tree_body, Tree):
        return False
    if len(tree_body.children) != 1:
        return False
    if not isinstance(tree_body.children[0], Tree):
        return False
    return tree_body.children[0].data == "object"


def _read_object_elems(tree_body: Tree) -> list[Tree]:
    object_elements = []
    for obj_child in tree_body.children[0].children:
        if not isinstance(obj_child, Tree):
            continue
        if obj_child.data != "object_elem":
            continue
        object_elements.append(obj_child)
    return object_elements


def _read_object_elem_key(tree_body: Tree) -> str:
    name_tree, _, _ = tree_body.children
    return token_name(name_tree.children[0])


def read_block_attribute_object_keys(tf_path: Path, block_type: str, block_name: str, block_key: str) -> list[str]:
    tree = safe_parse(tf_path)
    if tree is None:
        return []
    env_vars = []

    def extract_env_vars(tree: Tree) -> bool:
        if not is_block_type(tree, block_type):
            return False
        name, body = _block_name_body(tree)
        if name != block_name:
            return False
        attribute_value = _read_attribute(body, block_key)
        if not attribute_value:
            return False
        if not _is_object(attribute_value):
            return False
        object_elements = _read_object_elems(attribute_value)
        for obj_elem in object_elements:
            key = _read_object_elem_key(obj_elem)
            env_vars.append(key)
        return False

    def tree_call(tree: Tree) -> Tree:
        return tree

    process_generic(
        tree,
        extract_env_vars,
        tree_call,
    )
    return env_vars



from __future__ import annotations

import logging
import re
from collections import defaultdict
from typing import NamedTuple

from model_lib import Entity
from pydantic import Field

from atlas_init.cli_tf.schema_table_models import (
    AttrRefLine,
    FuncCallLine,
    TFSchemaAttribute,
)

logger = logging.getLogger(__name__)


def parse_attribute_ref(
    name: str, rest: str, go_code: str, code_lines: list[str], ref_line_nr: int
) -> TFSchemaAttribute | None:
    attr_ref = rest.lstrip("&").rstrip(",").strip()
    if not attr_ref.isidentifier():
        return None
    try:
        _instantiate_regex = re.compile(rf"{attr_ref}\s=\sschema\.\w+\{{$", re.M)
    except re.error:
        return None
    instantiate_match = _instantiate_regex.search(go_code)
    if not instantiate_match:
        return None
    line_start_nr = go_code[: instantiate_match.start()].count("\n") + 1
    line_start = code_lines[line_start_nr]
    attribute = parse_attribute_lines(code_lines, line_start_nr, line_start, name, is_attr_ref=True)
    attribute.attr_ref_line = AttrRefLine(line_nr=ref_line_nr, attr_ref=attr_ref)
    return attribute


def parse_func_call_line(
    name: str, rest: str, lines: list[str], go_code: str, call_line_nr: int
) -> TFSchemaAttribute | None:
    func_def_line = _function_line(rest, go_code)
    if not func_def_line:
        return None
    func_name, args = rest.split("(", maxsplit=1)
    func_start, func_end = _func_lines(name, lines, func_def_line)
    call = FuncCallLine(
        call_line_nr=call_line_nr,
        func_name=func_name.strip(),
        args=args.removesuffix("),").strip(),
        func_line_start=func_start,
        func_line_end=func_end,
    )
    return TFSchemaAttribute(
        name=name,
        lines=lines[func_start:func_end],
        line_start=func_start,
        line_end=func_end,
        func_call=call,
        indent="\t",
    )


def _func_lines(name: str, lines: list[str], func_def_line: str) -> tuple[int, int]:
    start_line = lines.index(func_def_line)
    for line_nr, line in enumerate(lines[start_line + 1 :], start=start_line + 1):
        if line.rstrip() == "}":
            return start_line, line_nr
    raise ValueError(f"no end line found for {name} on line {start_line}: {func_def_line}")


def _function_line(rest: str, go_code: str) -> str:
    function_name = rest.split("(")[0].strip()
    pattern = re.compile(rf"func {function_name}\(.*\) \*?schema\.\w+ \{{$", re.M)
    match = pattern.search(go_code)
    if not match:
        return ""
    return go_code[match.start() : match.end()]


def parse_attribute_lines(
    lines: list[str],
    line_nr: int,
    line: str,
    name: str,
    *,
    is_attr_ref: bool = False,
) -> TFSchemaAttribute:
    indents = len(line) - len(line.lstrip())
    indent = indents * "\t"
    end_line = f"{indent}}}" if is_attr_ref else f"{indent}}},"
    for extra_lines, next_line in enumerate(lines[line_nr + 1 :], start=1):
        if next_line == end_line:
            return TFSchemaAttribute(
                name=name,
                lines=lines[line_nr : line_nr + extra_lines],
                line_start=line_nr,
                line_end=line_nr + extra_lines,
                indent=indent,
            )
    raise ValueError(f"no end line found for {name}, starting on line {line_nr}")


_schema_attribute_go_regex = re.compile(
    r'^\s+"(?P<name>[^"]+)":\s(?P<rest>.+)$',
)


def find_attributes(go_code: str) -> list[TFSchemaAttribute]:
    lines = ["", *go_code.splitlines()]  # support line_nr indexing
    attributes = []
    for line_nr, line in enumerate(lines):
        match = _schema_attribute_go_regex.match(line)
        if not match:
            continue
        name = match.group("name")
        rest = match.group("rest")
        if rest.endswith("),"):
            if attr := parse_func_call_line(name, rest, lines, go_code, line_nr):
                attributes.append(attr)
        elif attr := parse_attribute_ref(name, rest, go_code, lines, line_nr):
            attributes.append(attr)
        else:
            try:
                attr = parse_attribute_lines(lines, line_nr, line, name)
            except ValueError as e:
                logger.warning(e)
                continue
            if not attr.type:
                continue
            attributes.append(attr)
    set_attribute_paths(attributes)
    return attributes


class StartEnd(NamedTuple):
    start: int
    end: int
    name: str
    func_call_line: FuncCallLine | None

    def has_parent(self, other: StartEnd) -> bool:
        if self.name == other.name:
            return False
        if func_call := self.func_call_line:
            func_call_line = func_call.call_line_nr
            return other.start < func_call_line < other.end
        return self.start > other.start and self.end < other.end


def set_attribute_paths(attributes: list[TFSchemaAttribute]) -> list[TFSchemaAttribute]:
    start_stops = [StartEnd(a.line_start, a.line_end, a.name, a.func_call) for a in attributes]
    overlaps = [
        (attribute, [other for other in start_stops if start_stop.has_parent(other)])
        for attribute, start_stop in zip(attributes, start_stops, strict=False)
    ]
    for attribute, others in overlaps:
        if not others:
            attribute.attribute_path = attribute.name
            continue
        overlaps = defaultdict(list)
        for other in others:
            overlaps[(other.start, other.end)].append(other.name)
        paths = []
        for names in overlaps.values():
            if len(names) == 1:
                paths.append(names[0])
            else:
                paths.append(f"({'|'.join(names)})")
        paths.append(attribute.name)
        attribute.attribute_path = ".".join(paths)
    return attributes


class GoSchemaFunc(Entity):
    name: str
    line_start: int
    line_end: int
    call_attributes: list[TFSchemaAttribute] = Field(default_factory=list)
    attributes: list[TFSchemaAttribute] = Field(default_factory=list)

    @property
    def attribute_names(self) -> set[str]:
        return {a.name for a in self.call_attributes}

    @property
    def attribute_paths(self) -> str:
        paths = set()
        for a in self.call_attributes:
            path = ".".join(a.parent_attribute_names())
            paths.add(path)
        return f"({'|'.join(paths)})" if len(paths) > 1 else paths.pop()

    def contains_attribute(self, attribute: TFSchemaAttribute) -> bool:
        names = self.attribute_names
        return any(parent_attribute in names for parent_attribute in attribute.parent_attribute_names())


def find_schema_functions(attributes: list[TFSchemaAttribute]) -> list[GoSchemaFunc]:
    function_call_attributes = defaultdict(list)
    for a in attributes:
        if a.is_function_call:
            call = a.func_call
            assert call
            function_call_attributes[call.func_name].append(a)
    root_function = GoSchemaFunc(name="", line_start=0, line_end=0)
    functions: list[GoSchemaFunc] = [
        GoSchemaFunc(
            name=name,
            line_start=func_attributes[0].line_start,
            line_end=func_attributes[0].line_end,
            call_attributes=func_attributes,
        )
        for name, func_attributes in function_call_attributes.items()
    ]
    for attribute in attributes:
        if match_functions := [func for func in functions if func.contains_attribute(attribute)]:
            func_names = [func.name for func in match_functions]
            err_msg = f"multiple functions found for {attribute.name}, {func_names}"
            assert len(match_functions) == 1, err_msg
            function = match_functions[0]
            function.attributes.append(attribute)
            attribute.absolute_attribute_path = f"{function.attribute_paths}.{attribute.attribute_path}".lstrip(".")
        else:
            root_function.attributes.append(attribute)
            attribute.absolute_attribute_path = attribute.attribute_path
    return [root_function, *functions]


def parse_schema_functions(
    go_code: str,
) -> tuple[list[TFSchemaAttribute], list[GoSchemaFunc]]:
    attributes = find_attributes(go_code)
    functions = find_schema_functions(attributes)
    return sorted(attributes), functions



import logging
import os

import typer
from zero_3rdparty.file_utils import clean_dir

from atlas_init.cli_cfn.aws import (
    activate_resource_type,
    deactivate_third_party_type,
    deregister_cfn_resource_type,
    get_last_cfn_type,
    wait_on_stack_ok,
)
from atlas_init.cli_cfn.aws import (
    delete_stack as delete_stack_aws,
)
from atlas_init.cli_cfn.contract import contract_test_cmd
from atlas_init.cli_cfn.example import example_cmd
from atlas_init.cli_cfn.files import (
    create_sample_file_from_input,
    has_md_link,
    iterate_schemas,
)
from atlas_init.cli_helper.run import run_command_is_ok
from atlas_init.cloud.aws import run_in_regions
from atlas_init.repos.cfn import (
    validate_type_name_regions,
)
from atlas_init.repos.path import Repo, current_dir, find_paths, resource_root
from atlas_init.settings.env_vars import active_suites, init_settings
from atlas_init.settings.env_vars_modules import TFModuleCfn

app = typer.Typer(no_args_is_help=True)
app.command(name="example")(example_cmd)
app.command(name="contract-test")(contract_test_cmd)
logger = logging.getLogger(__name__)


@app.command()
def reg(
    type_name: str,
    region_filter: str = typer.Option(default=""),
    dry_run: bool = typer.Option(False),
):
    if dry_run:
        logger.info("dry-run is set")
    type_name, regions = validate_type_name_regions(type_name, region_filter)
    assert len(regions) == 1, f"are you sure you want to activate {type_name} in all regions?"
    region = regions[0]
    found_third_party = deactivate_third_party_type(type_name, region, dry_run=dry_run)
    if not found_third_party:
        local = get_last_cfn_type(type_name, region, is_third_party=False)
        if local:
            deregister_cfn_resource_type(type_name, deregister=not dry_run, region_filter=region)
    logger.info(f"ready to activate {type_name}")
    init_settings(TFModuleCfn)
    cfn_execution_role = TFModuleCfn.from_env().CFN_EXAMPLE_EXECUTION_ROLE
    last_third_party = get_last_cfn_type(type_name, region, is_third_party=True)
    assert last_third_party, f"no 3rd party extension found for {type_name} in {region}"
    if dry_run:
        return
    activate_resource_type(last_third_party, region, cfn_execution_role)
    logger.info(f"{type_name} {last_third_party.version} is activated ✅")


@app.command()
def dereg(
    type_name: str,
    region_filter: str = typer.Option(default=""),
    dry_run: bool = typer.Option(False),
    is_local: bool = typer.Option(False),
):
    if dry_run:
        logger.info("dry-run is set")
    type_name, regions = validate_type_name_regions(type_name, region_filter)

    def deactivate(region: str):
        deactivate_third_party_type(type_name, region, dry_run=dry_run)

    def deactivate_local(region: str):
        deregister_cfn_resource_type(type_name, deregister=True, region_filter=region)

    if is_local:
        logger.info("deregistering local")
        run_in_regions(deactivate_local, regions)
    else:
        logger.info("deregistering 3rd party")
        run_in_regions(deactivate, regions)


@app.command()
def inputs(
    context: typer.Context,
    skip_samples: bool = typer.Option(default=False),
    single_input: int = typer.Option(0, "--input", "-i", help="keep only input_X files"),
):
    settings = init_settings()
    suites = active_suites(settings)
    assert len(suites) == 1, "no test suit found"
    cwd = current_dir()
    suite = suites[0]
    assert suite.cwd_is_repo_go_pkg(cwd, repo_alias="cfn")
    CREATE_FILENAME = "cfn-test-create-inputs.sh"  # noqa: N806
    create_dirs = ["test/contract-testing", "test"]
    parent_dir = None
    for parent in create_dirs:
        parent_candidate = cwd / parent / CREATE_FILENAME
        if parent_candidate.exists():
            parent_dir = parent
            break
    assert parent_dir, f"unable to find a {CREATE_FILENAME} in {create_dirs} in {cwd}"
    if not run_command_is_ok(
        cwd=cwd,
        cmd=f"./{parent_dir}/{CREATE_FILENAME}" + " ".join(context.args),
        env={**os.environ},
        logger=logger,
    ):
        logger.critical("failed to create cfn contract input files")
        raise typer.Exit(1)
    inputs_dir = cwd / "inputs"
    samples_dir = cwd / "samples"
    log_group_name = f"mongodb-atlas-{cwd.name}-logs"
    if not skip_samples and samples_dir.exists():
        clean_dir(samples_dir)
    expected_input = ""
    if single_input:
        logger.warning(f"will only use input_{single_input}")
        expected_input = f"inputs_{single_input}"
    for file in sorted(inputs_dir.glob("*.json")):
        if single_input and not file.name.startswith(expected_input):
            file.unlink()
            continue
        logger.info(f"input exist at inputs/{file.name} ✅")
        if skip_samples:
            continue
        create_sample_file_from_input(samples_dir, log_group_name, file)
    if single_input:
        for file in sorted(inputs_dir.glob("*.json")):
            new_name = file.name.replace(expected_input, "inputs_1")
            new_filename = inputs_dir / new_name
            file.rename(new_filename)
            logger.info(f"renamed from {file} -> {new_filename}")


@app.command()
def gen_docs():
    repo_path, *_ = find_paths(Repo.CFN)
    root = resource_root(repo_path)
    for path, schema in iterate_schemas(root):
        if has_md_link(schema.description):
            logger.warning(f"found md link in {schema.type_name} in {path}")


@app.command()
def wait_on_stack(
    stack_name: str = typer.Argument(...),
    region: str = typer.Argument(...),
    timeout_s: int = typer.Option(300, "-t", "--timeout-seconds"),
):
    wait_on_stack_ok(stack_name, region, timeout_seconds=timeout_s)
    logger.info(f"stack {stack_name} in {region} is ready ✅")


@app.command()
def delete_stack(
    stack_name: str = typer.Argument(...),
    region: str = typer.Argument(...),
):
    delete_stack_aws(region, stack_name)
    logger.info(f"stack {stack_name} in {region} is deleted ✅")



from pathlib import Path

VERSION = "0.7.0"


def running_in_repo() -> bool:
    git_config = Path(__file__).parent.parent / ".git/config"
    return git_config.exists() and "atlas-init" in git_config.read_text()



from __future__ import annotations
from typing import Self

from model_lib import Entity
from pydantic import Field, RootModel, model_validator

from atlas_init.tf_ext.tf_dep import AtlasGraph

_emojii_list = [
    "1️⃣",
    "2️⃣",
    "3️⃣",
    "4️⃣",
    "5️⃣",
    "6️⃣",
    "7️⃣",
    "8️⃣",
    "9️⃣",
    "🔟",
    "1️⃣1️⃣",
    "1️⃣2️⃣",
]
_emoji_counter = 0


def choose_next_emoji() -> str:
    global _emoji_counter
    emoji = _emojii_list[_emoji_counter]
    _emoji_counter += 1
    return emoji


class ModuleState(Entity):
    resource_types: set[str] = Field(default_factory=set, description="Set of resource types in the module.")


def default_allowed_multi_parents() -> set[str]:
    return {
        "mongodbatlas_project",
    }


class ModuleConfig(Entity):
    name: str = Field(..., description="Name of the module.")
    root_resource_types: list[str] = Field(..., description="List of root resource types for the module.")
    force_include_children: list[str] = Field(
        default_factory=list, description="List of resource types that should always be included as children."
    )
    emojii: str = Field(init=False, default_factory=choose_next_emoji)
    allowed_multi_parents: set[str] = Field(
        default_factory=default_allowed_multi_parents,
        description="Set of parents that a child resource type can have in addition to the root_resource_type.",
    )
    allow_external_dependencies: bool = Field(
        default=False, description="Whether to allow external dependencies for the module."
    )
    extra_nested_resource_types: list[str] = Field(
        default_factory=list,
        description="List of additional nested resource types that should be included in the module.",
    )

    state: ModuleState = Field(default_factory=ModuleState, description="Internal state of the module.")

    @model_validator(mode="after")
    def update_state(self) -> Self:
        self.state.resource_types.update(self.root_resource_types)
        return self

    @property
    def tree_label(self) -> str:
        return f"{self.emojii} {self.name}"

    def include_child(self, child: str, atlas_graph: AtlasGraph) -> bool:
        if child in atlas_graph.deprecated_resource_types:
            return False
        if child in self.force_include_children or child in self.extra_nested_resource_types:
            self.state.resource_types.add(child)
            return True
        has_external_dependencies = len(atlas_graph.external_parents.get(child, [])) > 0
        if self.allow_external_dependencies and has_external_dependencies:
            has_external_dependencies = False
        is_a_parent = bool(atlas_graph.parent_child_edges.get(child))
        extra_parents = (
            set(atlas_graph.all_parents(child))
            - self.allowed_multi_parents
            - set(self.root_resource_types)
            - set(self.extra_nested_resource_types)
        )
        has_extra_parents = len(extra_parents) > 0
        if has_external_dependencies or is_a_parent or has_extra_parents:
            return False
        self.state.resource_types.add(child)
        return True


class ModuleConfigs(RootModel[dict[str, ModuleConfig]]):
    def module_emoji_prefix(self, resource_type: str) -> str:
        """Get the emoji prefix for a resource type based on its module."""
        return next(
            (
                module_config.emojii
                for module_config in self.root.values()
                if resource_type in module_config.state.resource_types
            ),
            "",
        )



from __future__ import annotations

import re
from dataclasses import dataclass
from enum import StrEnum
from functools import total_ordering
from typing import ClassVar, Literal, NamedTuple, Self, TypeAlias

import humanize
from model_lib import Entity, utc_datetime_ms
from pydantic import Field, model_validator
from zero_3rdparty import iter_utils
from zero_3rdparty.datetime_utils import utc_now
from zero_3rdparty.str_utils import instance_repr

from atlas_init.cli_tf.go_test_run import GoTestRun
from atlas_init.repos.go_sdk import ApiSpecPaths


class GoTestErrorClass(StrEnum):
    """Goal of each error class to be actionable."""

    FLAKY_400 = "flaky_400"
    FLAKY_500 = "flaky_500"
    FLAKY_CHECK = "flaky_check"
    FLAKY_CLIENT = "flaky_client"
    OUT_OF_CAPACITY = "out_of_capacity"
    PROJECT_LIMIT_EXCEEDED = "project_limit_exceeded"
    DANGLING_RESOURCE = "dangling_resource"
    REAL_TEST_FAILURE = "real_test_failure"
    TIMEOUT = "timeout"
    UNKNOWN = "unknown"
    PROVIDER_DOWNLOAD = "provider_download"
    UNCLASSIFIED = "unclassified"

    __ACTIONS__ = {
        FLAKY_400: "retry",
        FLAKY_500: "retry",
        FLAKY_CHECK: "retry",
        FLAKY_CLIENT: "retry",
        PROVIDER_DOWNLOAD: "retry",
        OUT_OF_CAPACITY: "retry_later",
        PROJECT_LIMIT_EXCEEDED: "clean_project",
        DANGLING_RESOURCE: "update_cleanup_script",
        REAL_TEST_FAILURE: "investigate",
        TIMEOUT: "investigate",
        UNKNOWN: "investigate",
    }
    __CONTAINS_MAPPING__ = {
        OUT_OF_CAPACITY: ("OUT_OF_CAPACITY",),
        FLAKY_500: ("HTTP 500", "UNEXPECTED_ERROR", "503 Service Unavailable"),
        FLAKY_CLIENT: ("dial tcp: lookup", "i/o timeout"),
        PROVIDER_DOWNLOAD: [
            "mongodbatlas: failed to retrieve authentication checksums for provider",
            "Error: Failed to install provider github.com: bad response",
        ],
        TIMEOUT: ("timeout while waiting for",),
    }

    @classmethod
    def auto_classification(cls, output: str) -> GoTestErrorClass | None:
        def contains(output: str, contains_part: str) -> bool:
            if " " in contains_part:
                return all(part in output for part in contains_part.split())
            return contains_part in output

        return next(
            (
                error_class
                for error_class, contains_list in cls.__CONTAINS_MAPPING__.items()
                if any(contains(output, contains_part) for contains_part in contains_list)
            ),
            None,
        )  # type: ignore


API_METHODS = ["GET", "POST", "PUT", "DELETE", "PATCH"]


class GoTestAPIError(Entity):
    type: Literal["api_error"] = "api_error"
    api_error_code_str: str
    api_path: str
    api_method: Literal["GET", "POST", "PUT", "DELETE", "PATCH"]
    api_response_code: int
    tf_resource_name: str = ""
    tf_resource_type: str = ""
    step_nr: int = -1

    api_path_normalized: str = Field(init=False, default="")

    @model_validator(mode="after")
    def strip_path_chars(self) -> GoTestAPIError:
        self.api_path = self.api_path.rstrip(":/")
        return self

    def add_info_fields(self, info: DetailsInfo) -> None:
        if api_paths := info.paths:
            self.api_path_normalized = api_paths.normalize_path(self.api_method, self.api_path)

    def __str__(self) -> str:
        resource_part = f"{self.tf_resource_type} " if self.tf_resource_type else ""
        if self.api_path_normalized:
            return f"{resource_part}{self.api_error_code_str} {self.api_method} {self.api_path_normalized} {self.api_response_code}"
        return f"{resource_part}{self.api_error_code_str} {self.api_method} {self.api_path} {self.api_response_code}"


@total_ordering
class CheckError(Entity):
    attribute: str = ""
    expected: str = ""
    got: str = ""
    check_nr: int = -1

    def __lt__(self, other) -> bool:
        if not isinstance(other, CheckError):
            raise TypeError
        return (self.check_nr, self.attribute) < (other.check_nr, other.attribute)

    def __str__(self) -> str:
        if self.attribute and self.expected and self.got:
            return f"{self.check_nr}({self.attribute}:expected:{self.expected}, got: {self.got})"
        return f"{self.check_nr}"

    @classmethod
    def parse_from_output(cls, output: str) -> list[Self]:
        return [
            cls(**check_match.groupdict())  # type: ignore
            for check_match in check_pattern.finditer(output)
        ]


class GoTestResourceCheckError(Entity):
    type: Literal["check_error"] = "check_error"
    tf_resource_name: str
    tf_resource_type: str
    step_nr: int = -1
    check_errors: list[CheckError] = Field(default_factory=list)
    test_name: str = ""

    def add_info_fields(self, info: DetailsInfo) -> None:
        self.test_name = info.run.name

    def __str__(self) -> str:
        return f"{self.tf_resource_type} {self.tf_resource_name} {self.step_nr} {self.check_errors}"

    @property
    def check_numbers_str(self) -> str:
        return ",".join(str(check.check_nr) for check in sorted(self.check_errors))

    def check_errors_match(self, other_check_errors: list[CheckError]) -> bool:
        if len(self.check_errors) != len(other_check_errors):
            return False
        return all(
            any(
                check.check_nr == other_check.check_nr and check.attribute == other_check.attribute
                for other_check in other_check_errors
            )
            for check in self.check_errors
        )


class GoTestGeneralCheckError(Entity):
    type: Literal["general_check_error"] = "general_check_error"
    step_nr: int = -1
    check_errors: list[CheckError] = Field(default_factory=list)
    error_check_str: str
    test_name: str = ""

    def add_info_fields(self, info: DetailsInfo) -> None:
        self.test_name = info.run.name

    def check_errors_str(self) -> str:
        return ",".join(str(check) for check in sorted(self.check_errors))

    def __str__(self) -> str:
        return f"Step {self.step_nr} {self.check_errors_str()}"


@dataclass
class DetailsInfo:
    run: GoTestRun
    paths: ApiSpecPaths | None = None


class GoTestDefaultError(Entity):
    type: Literal["default_error"] = "default_error"
    error_str: str

    def add_info_fields(self, _: DetailsInfo) -> None:
        pass


ErrorDetailsT: TypeAlias = GoTestAPIError | GoTestResourceCheckError | GoTestDefaultError | GoTestGeneralCheckError


class ErrorClassified(NamedTuple):
    classified: dict[GoTestErrorClass, list[GoTestError]]
    unclassified: list[GoTestError]


class ErrorClassAuthor(StrEnum):
    AUTO = "auto"
    HUMAN = "human"
    LLM = "llm"
    SIMILAR = "similar"


class GoTestErrorClassification(Entity):
    error_class: GoTestErrorClass = GoTestErrorClass.UNCLASSIFIED
    ts: utc_datetime_ms = Field(default_factory=utc_now)
    author: ErrorClassAuthor
    confidence: float = 0.0
    test_output: str = ""
    details: ErrorDetailsT
    run_id: str
    test_name: str

    STR_COLUMNS: ClassVar[list[str]] = ["error_class", "author", "run_id", "confidence", "ts_when"]

    def needs_classification(self, confidence_threshold: float = 1.0) -> bool:
        return (
            self.error_class in {GoTestErrorClass.UNCLASSIFIED, GoTestErrorClass.UNKNOWN}
            or self.confidence < confidence_threshold
        )

    @property
    def ts_when(self) -> str:
        return humanize.naturaltime(self.ts)

    def __str__(self) -> str:
        return instance_repr(self, self.STR_COLUMNS)


@total_ordering
class GoTestError(Entity):
    details: ErrorDetailsT
    run: GoTestRun
    bot_error_class: GoTestErrorClass = GoTestErrorClass.UNCLASSIFIED
    human_error_class: GoTestErrorClass = GoTestErrorClass.UNCLASSIFIED

    def __lt__(self, other) -> bool:
        if not isinstance(other, GoTestError):
            raise TypeError
        return self.run < other.run

    @property
    def run_id(self) -> str:
        return self.run.id

    @property
    def run_name(self) -> str:
        return self.run.name

    @property
    def classifications(self) -> tuple[GoTestErrorClass, GoTestErrorClass] | None:
        if (
            self.bot_error_class != GoTestErrorClass.UNCLASSIFIED
            and self.human_error_class != GoTestErrorClass.UNCLASSIFIED
        ):
            return self.bot_error_class, self.human_error_class
        return None

    def set_human_and_bot_classification(self, chosen_class: GoTestErrorClass) -> None:
        self.human_error_class = chosen_class
        self.bot_error_class = chosen_class

    def match(self, other: GoTestError) -> bool:
        if self.run.id == other.run.id:
            return True
        details = self.details
        other_details = other.details
        if type(self.details) is not type(other_details):
            return False
        if isinstance(details, GoTestAPIError):
            assert isinstance(other_details, GoTestAPIError)
            return (
                details.api_path_normalized == other_details.api_path_normalized
                and details.api_response_code == other_details.api_response_code
                and details.api_method == other_details.api_method
                and details.api_response_code == other_details.api_response_code
            )
        if isinstance(details, GoTestResourceCheckError):
            assert isinstance(other_details, GoTestResourceCheckError)
            return (
                details.tf_resource_name == other_details.tf_resource_name
                and details.tf_resource_type == other_details.tf_resource_type
                and details.step_nr == other_details.step_nr
                and details.check_numbers_str == other_details.check_numbers_str
            )
        return False

    @classmethod
    def group_by_classification(
        cls, errors: list[GoTestError], *, classifier: Literal["bot", "human"] = "human"
    ) -> ErrorClassified:
        def get_classification(error: GoTestError) -> GoTestErrorClass:
            if classifier == "bot":
                return error.bot_error_class
            return error.human_error_class

        grouped_errors: dict[GoTestErrorClass, list[GoTestError]] = iter_utils.group_by_once(
            errors, key=get_classification
        )
        unclassified = grouped_errors.pop(GoTestErrorClass.UNCLASSIFIED, [])
        return ErrorClassified(grouped_errors, unclassified)

    @classmethod
    def group_by_name_with_package(cls, errors: list[GoTestError]) -> dict[str, list[GoTestError]]:
        def by_name(error: GoTestError) -> str:
            return error.run.name_with_package

        return iter_utils.group_by_once(errors, key=by_name)

    @property
    def short_description(self) -> str:
        details = self.details
        return details_short_description(details) if details else ""

    def header(self, use_ticks: bool = False) -> str:
        name_with_ticks = f"`{self.run.name_with_package}`" if use_ticks else self.run.name_with_package
        if details := self.short_description:
            return f"{name_with_ticks} {details}"
        return f"{name_with_ticks}"


def details_short_description(details: ErrorDetailsT) -> str:
    match details:
        case GoTestGeneralCheckError():
            return str(details)
        case GoTestResourceCheckError():
            return f"CheckFailure for {details.tf_resource_type}.{details.tf_resource_name} at Step: {details.step_nr} Checks: {details.check_numbers_str}"
        case GoTestAPIError(api_path_normalized=api_path_normalized) if api_path_normalized:
            return f"API Error {details.api_error_code_str} {api_path_normalized}"
        case GoTestAPIError(api_path=api_path):
            return f"{details.api_error_code_str} {api_path}"
    return ""


one_of_methods = "|".join(API_METHODS)


check_pattern_str = r"Check (?P<check_nr>\d+)/\d+"
check_pattern = re.compile(check_pattern_str)
url_pattern = r"https://cloud(-dev|-qa)?\.mongodb\.com(?P<api_path>\S+)"
error_check_pattern = re.compile(check_pattern_str + r"\s+error:\s(?P<error_check_str>.+)$", re.MULTILINE)
detail_patterns: list[re.Pattern] = [
    re.compile(r"Step (?P<step_nr>\d+)/\d+"),
    check_pattern,
    re.compile(r"mongodbatlas_(?P<tf_resource_type>[^\.]+)\.(?P<tf_resource_name>[\w_-]+)"),
    re.compile(rf"(?P<api_method>{one_of_methods})" + r": HTTP (?P<api_response_code>\d+)"),
    re.compile(r'Error code: "(?P<api_error_code_str>[^"]+)"'),
    re.compile(url_pattern),
]

# Error: error creating MongoDB Cluster: POST https://cloud-dev.mongodb.com/api/atlas/v1.0/groups/680ecbc7122f5b15cc627ba5/clusters: 409 (request "OUT_OF_CAPACITY") The requested region is currently out of capacity for the requested instance size.
api_error_pattern_missing_details = re.compile(
    rf"(?P<api_method>{one_of_methods})\s+"
    + url_pattern
    + r'\s+(?P<api_response_code>\d+)\s\(request\s"(?P<api_error_code_str>[^"]+)"\)'
)


def parse_error_details(run: GoTestRun) -> ErrorDetailsT:
    kwargs = {}
    output = run.output_lines_str
    for pattern in detail_patterns:
        if pattern_match := pattern.search(output):
            kwargs |= pattern_match.groupdict()
    match kwargs:
        case {"api_path": _, "api_error_code_str": _}:
            return GoTestAPIError(**kwargs)
        case {"api_path": _} if pattern_match := api_error_pattern_missing_details.search(output):
            kwargs |= pattern_match.groupdict()
            return GoTestAPIError(**kwargs)
        case {"check_nr": _} if all(name in kwargs for name in ("tf_resource_name", "tf_resource_type")):
            kwargs.pop("check_nr")
            check_errors = CheckError.parse_from_output(output)
            return GoTestResourceCheckError(**kwargs, check_errors=check_errors)
        case {"check_nr": _}:
            if error_check_match := error_check_pattern.search(output):
                kwargs.pop("check_nr")
                check_errors = CheckError.parse_from_output(output)
                return GoTestGeneralCheckError(
                    **kwargs, error_check_str=error_check_match.group("error_check_str"), check_errors=check_errors
                )
    kwargs.pop("error_check_str", None)  # Remove if it was not matched
    return GoTestDefaultError(error_str=run.output_lines_str)



from __future__ import annotations

import logging
import os
from contextlib import suppress
from functools import cached_property
from pathlib import Path
from typing import Any, NamedTuple, TypeVar

from model_lib import StaticSettings, parse_payload
from pydantic import BaseModel, ValidationError, field_validator
from zero_3rdparty import iter_utils, str_utils

from atlas_init.settings.config import (
    AtlasInitConfig,
    TestSuite,
)
from atlas_init.settings.config import (
    active_suites as config_active_suites,
)
from atlas_init.settings.path import (
    DEFAULT_ATLAS_INIT_CONFIG_PATH,
    DEFAULT_ATLAS_INIT_SCHEMA_CONFIG_PATH,
    DEFAULT_TF_SRC_PATH,
    dump_dotenv,
    load_dotenv,
    repo_path_rel_path,
)

logger = logging.getLogger(__name__)
ENV_PREFIX = "ATLAS_INIT_"
DEFAULT_PROFILE = "default"
DEFAULT_PROJECT_NAME = "atlas-init"
ENV_S3_PROFILE_BUCKET = f"{ENV_PREFIX}S3_PROFILE_BUCKET"
ENV_PROJECT_NAME = f"{ENV_PREFIX}PROJECT_NAME"
ENV_PROFILE = f"{ENV_PREFIX}PROFILE"
ENV_TEST_SUITES = f"{ENV_PREFIX}TEST_SUITES"
ENV_CLIPBOARD_COPY = f"{ENV_PREFIX}CLIPBOARD_COPY"
FILENAME_ENV_MANUAL = ".env-manual"
T = TypeVar("T")


def read_from_env(env_key: str, default: str = "") -> str:
    return next(
        (os.environ[name] for name in [env_key, env_key.lower(), env_key.upper()] if name in os.environ),
        default,
    )


class AtlasInitSettings(StaticSettings):
    atlas_init_profile: str = DEFAULT_PROFILE  # override this for different env, e.g. dev, prod
    atlas_init_project_name: str = DEFAULT_PROJECT_NAME  # used in the atlas cloud
    atlas_init_config_path: Path = DEFAULT_ATLAS_INIT_CONFIG_PATH  # /atlas_init.yaml
    atlas_init_tf_src_path: Path = DEFAULT_TF_SRC_PATH  # /tf directory of repo
    atlas_init_tf_schema_config_path: Path = DEFAULT_ATLAS_INIT_SCHEMA_CONFIG_PATH  # /terraform.yaml
    atlas_init_schema_out_path: Path | None = None  # override this for the generated schema
    atlas_init_static_html_path: Path | None = None

    atlas_init_cfn_profile: str = ""
    atlas_init_cfn_region: str = ""
    atlas_init_cfn_use_kms_key: bool = False
    atlas_init_cliboard_copy: str = ""
    atlas_init_test_suites: str = ""
    atlas_init_s3_profile_bucket: str = ""

    non_interactive: bool = False

    mongo_database: str = "atlas_init"
    mongo_url: str = "mongodb://user:pass@localhost:27017?retryWrites=true&w=majority&authSource=admin"

    @property
    def is_interactive(self) -> bool:
        return not self.non_interactive

    @property
    def profiles_path(self) -> Path:
        return self.static_root / "profiles"

    @property
    def project_name(self) -> str:
        return self.atlas_init_project_name

    @property
    def profile(self) -> str:
        return self.atlas_init_profile

    @property
    def schema_out_path_computed(self) -> Path:
        return self.atlas_init_schema_out_path or self.static_root / "schema"

    @property
    def profile_dir(self) -> Path:
        return self.profiles_path / self.profile

    @property
    def env_file_manual(self) -> Path:
        return self.profile_dir / FILENAME_ENV_MANUAL

    @property
    def manual_env_vars(self) -> dict[str, str]:
        env_manual_path = self.env_file_manual
        return load_dotenv(env_manual_path) if env_manual_path.exists() else {}

    @property
    def env_vars_generated(self) -> Path:
        return self.profile_dir / ".env-generated"

    @property
    def env_vars_vs_code(self) -> Path:
        return self.profile_dir / ".env-vscode"

    @property
    def env_vars_trigger(self) -> Path:
        return self.profile_dir / ".env-trigger"

    @property
    def tf_data_dir(self) -> Path:
        return self.profile_dir / ".terraform"

    @property
    def tf_vars_path(self) -> Path:
        return self.tf_data_dir / "vars.auto.tfvars.json"

    @property
    def tf_state_path(self) -> Path:
        return self.profile_dir / "tf_state"

    @property
    def tf_outputs_path(self) -> Path:
        return self.profile_dir / "tf_outputs.json"

    @property
    def github_ci_run_logs(self) -> Path:
        return self.cache_root / "github_ci_run_logs"

    @property
    def github_ci_summary_dir(self) -> Path:
        return self.cache_root / "github_ci_summary"

    def github_ci_summary_path(self, summary_name: str) -> Path:
        return self.github_ci_summary_dir / str_utils.ensure_suffix(summary_name, ".md")

    def github_ci_summary_details_path(self, summary_name: str, test_name: str) -> Path:
        return self.github_ci_summary_path(summary_name).parent / self.github_ci_summary_details_rel_path(
            summary_name, test_name
        )

    def github_ci_summary_details_rel_path(self, summary_name: str, test_name: str) -> str:
        return f"{summary_name.removesuffix('.md')}_details/{test_name}.md"

    @property
    def go_test_logs_dir(self) -> Path:
        return self.cache_root / "go_test_logs"

    @property
    def atlas_atlas_api_transformed_yaml(self) -> Path:
        return self.cache_root / "atlas_api_transformed.yaml"

    def cfn_region(self, default: str) -> str:
        return self.atlas_init_cfn_region or default

    def include_extra_env_vars_in_vscode(self, extra_env_vars: dict[str, str]) -> None:
        extra_name = ", ".join(extra_env_vars.keys())
        original_env_vars = load_dotenv(self.env_vars_vs_code)
        new_env_vars = original_env_vars | extra_env_vars
        dump_dotenv(self.env_vars_vs_code, new_env_vars)
        logger.info(f"done {self.env_vars_vs_code} updated with {extra_name} env-vars ✅")

    @field_validator(ENV_TEST_SUITES.lower(), mode="after")
    @classmethod
    def ensure_whitespace_replaced_with_commas(cls, value: str) -> str:
        return value.strip().replace(" ", ",")

    @cached_property
    def config(self) -> AtlasInitConfig:
        config_path = (
            Path(self.atlas_init_config_path) if self.atlas_init_config_path else DEFAULT_ATLAS_INIT_CONFIG_PATH
        )
        assert config_path.exists(), f"no config path found @ {config_path}"
        yaml_parsed = parse_payload(config_path)
        assert isinstance(yaml_parsed, dict), f"config must be a dictionary, got {yaml_parsed}"
        return AtlasInitConfig(**yaml_parsed)

    @property
    def test_suites_parsed(self) -> list[str]:
        return [t for t in self.atlas_init_test_suites.split(",") if t]

    def tf_vars(self, default_aws_region: str) -> dict[str, Any]:
        variables = {}
        if self.atlas_init_cfn_profile:
            variables["cfn_config"] = {
                "profile": self.atlas_init_cfn_profile,
                "region": self.atlas_init_cfn_region or default_aws_region,
                "use_kms_key": self.atlas_init_cfn_use_kms_key,
            }
        if self.atlas_init_s3_profile_bucket:
            variables["use_aws_s3"] = True
        return variables


class EnvVarsCheck(NamedTuple):
    missing: list[str]
    ambiguous: list[str]

    @property
    def is_ok(self) -> bool:
        return not self.missing and not self.ambiguous


def active_suites(settings: AtlasInitSettings) -> list[TestSuite]:  # type: ignore
    repo_path, cwd_rel_path = repo_path_rel_path()
    return config_active_suites(settings.config, repo_path, cwd_rel_path, settings.test_suites_parsed)


class EnvVarsError(Exception):
    def __init__(self, missing: list[str], ambiguous: list[str]):
        self.missing = missing
        self.ambiguous = ambiguous
        super().__init__(f"missing: {missing}, ambiguous: {ambiguous}")

    def __str__(self) -> str:
        return f"missing: {self.missing}, ambiguous: {self.ambiguous}"


def collect_required_env_vars(settings_classes: list[type[BaseModel]]) -> list[str]:
    cls_required_env_vars: dict[str, list[str]] = {}
    for cls in settings_classes:
        try:
            cls()
        except ValidationError as error:
            cls_required_env_vars[cls.__name__] = [".".join(str(loc) for loc in e["loc"]) for e in error.errors()]
    return list(iter_utils.flat_map(cls_required_env_vars.values()))


def detect_ambiguous_env_vars(manual_env_vars: dict[str, str]) -> list[str]:
    ambiguous: list[str] = []
    for env_name, manual_value in manual_env_vars.items():
        env_value = read_from_env(env_name)
        if env_value and manual_value != env_value:
            ambiguous.append(env_name)
    return ambiguous


def find_missing_env_vars(required_env_vars: list[str], manual_env_vars: dict[str, str]) -> list[str]:
    return sorted(
        env_name for env_name in required_env_vars if read_from_env(env_name) == "" and env_name not in manual_env_vars
    )


def init_settings(
    *settings_classes: type[BaseModel],
    skip_ambiguous_check: bool = False,
) -> AtlasInitSettings:
    settings = AtlasInitSettings.from_env()
    profile_env_vars = settings.manual_env_vars
    vscode_env_vars = settings.env_vars_vs_code
    if vscode_env_vars.exists():
        profile_env_vars |= load_dotenv(vscode_env_vars)
    required_env_vars = collect_required_env_vars(list(settings_classes))
    ambiguous = [] if skip_ambiguous_check else detect_ambiguous_env_vars(profile_env_vars)
    missing_env_vars = find_missing_env_vars(required_env_vars, profile_env_vars)

    if ambiguous:
        logger.warning(
            f"ambiguous env_vars: {ambiguous} (specified both in cli/env & in .env-(manual|vscode) file with different values)"
        )
    if missing_env_vars or ambiguous:
        raise EnvVarsError(missing_env_vars, ambiguous)

    if new_updates := {k: v for k, v in profile_env_vars.items() if k not in os.environ}:
        logger.info(f"loading env-vars {','.join(sorted(new_updates))}")
        os.environ |= new_updates
    for cls in settings_classes:
        cls()  # ensure any errors are raised
    return AtlasInitSettings.from_env()


def env_vars_cls_or_none(t: type[T], *, dotenv_path: Path | None = None) -> T | None:
    explicit_vars: dict[str, str] = {}
    if dotenv_path and dotenv_path.exists():
        explicit_vars = load_dotenv(dotenv_path)
    with suppress(ValidationError):
        return t(**explicit_vars)



import re
from collections.abc import Iterable
from pathlib import Path
from typing import Annotated, Literal, NamedTuple

import requests
from model_lib import Event
from pydantic import StringConstraints


class Change(NamedTuple):
    change_type: Literal["changed from", "removed"]
    class_name: str
    func_name: str


regex = re.compile(r"\-\s\(?\\?\*?(?P<class_name>\w+)\)?\.(?P<func_name>\w+):\s(?P<change_type>(removed|changed from))")


def parse_line(line: str) -> Change | None:
    if match := regex.match(line):
        # change = match.groupdict()["change"]
        # func_name = match.groupdict()["func_name"]
        return Change(**match.groupdict())  # type: ignore
    return None


class BreakingChange(NamedTuple):
    version: str
    line: str


def parse_breaking_changes(
    breaking_change_dir: Path,
    start_sdk_version: str = "v20231115000",
    end_sdk_version: str = "v20991115000",
) -> dict[Change, BreakingChange]:
    changes: dict[Change, BreakingChange] = {}
    for file in breaking_change_dir.glob("*.md"):
        # v20231115009.md
        sdk_version = file.name.removesuffix(".md")
        if sdk_version <= start_sdk_version or sdk_version > end_sdk_version:
            continue
        for line in file.read_text().splitlines():
            if change := parse_line(line):
                changes[change] = BreakingChange(sdk_version, line)
    return changes


def find_breaking_changes(text: str, changes: dict[Change, BreakingChange]) -> dict[Change, BreakingChange]:
    found_changes: dict[Change, BreakingChange] = {}
    for change, change_value in changes.items():
        if change.class_name in text and change.func_name in text:
            found_changes[change] = change_value
    return found_changes


def format_breaking_changes(text: str, changes: dict[Change, BreakingChange]):
    warning: list[str] = []
    for (_, cls_name, func_name), breaking_change in changes.items():
        warning.append(f"## {breaking_change.version}: {breaking_change.line}")
        for line_nr, line in enumerate(text.splitlines(), 1):
            if cls_name in line:
                warning.append(f"L{line_nr:03}: '{cls_name}' {line}")
            if func_name in line:
                warning.append(f"L{line_nr:03}: '{func_name}' {line}")
        warning.append("")
    return "\n".join(warning)


def is_removed(changes: Iterable[Change]) -> bool:
    return any(change.change_type == "removed" for change in changes)


def find_latest_sdk_version() -> str:
    response = requests.get("https://api.github.com/repos/mongodb/atlas-sdk-go/releases/latest", timeout=10)
    response.raise_for_status()
    name = response.json()["name"]
    assert isinstance(name, str)
    return name.removesuffix(".0.0")


SdkVersion = Annotated[str, StringConstraints(pattern="v\\d{11}")]


class SdkVersionUpgrade(Event):
    old: SdkVersion
    new: SdkVersion


SDK_VERSION_HELP = "e.g., v20231115008 in go.mongodb.org/atlas-sdk/XXXX/admin"



from __future__ import annotations

import re
from enum import StrEnum
from functools import total_ordering

from model_lib import Entity, Event
from pydantic import Field

from atlas_init.cli_tf.schema_v3 import ComputedOptionalRequired


class TFSchemaTableColumn(StrEnum):
    Computability = "Computability"
    Type = "Type"
    Default = "Default"
    PlanModifiers = "PlanModifiers"
    Deprecated = "Deprecated"


class FuncCallLine(Event):
    call_line_nr: int
    func_name: str
    args: str
    func_line_start: int
    func_line_end: int


class AttrRefLine(Event):
    line_nr: int
    attr_ref: str


_schema_type_regex = re.compile(r"schema\.\w+")


@total_ordering
class TFSchemaAttribute(Entity):
    name: str
    default: str = ""
    plan_modifiers: list[str] = Field(default_factory=list)

    line_start: int
    line_end: int
    indent: str = ""
    lines: list[str] = Field(default_factory=list)

    func_call: FuncCallLine | None = None
    attr_ref_line: AttrRefLine | None = None
    attribute_path: str = ""
    absolute_attribute_path: str = ""

    @property
    def type(self) -> str:
        if self.lines and (type_match := _schema_type_regex.search(self.lines[0])):
            return type_match.group()
        prefix = f"{self.indent}\tType:"
        for line in self.lines:
            if line.startswith(prefix):
                schema_type = line[len(prefix) :].strip()
                if schema_type.startswith("schema."):
                    return schema_type
        return ""

    @property
    def computability(self) -> ComputedOptionalRequired:
        if self.is_required:
            return ComputedOptionalRequired.required
        optional = self.is_optional
        computed = self.is_computed
        if optional:
            if computed:
                return ComputedOptionalRequired.computed_optional
            return ComputedOptionalRequired.optional
        if computed:
            return ComputedOptionalRequired.computed
        return ComputedOptionalRequired.unset

    @property
    def is_function_call(self) -> bool:
        return self.func_call is not None

    @property
    def is_required(self) -> bool:
        prefix = f"{self.indent}\tRequired:"
        return any("true" in line and line.startswith(prefix) for line in self.lines)

    @property
    def is_computed(self) -> bool:
        prefix = f"{self.indent}\tComputed:"
        return any("true" in line and line.startswith(prefix) for line in self.lines)

    @property
    def is_optional(self) -> bool:
        prefix = f"{self.indent}\tOptional:"
        return any("true" in line and line.startswith(prefix) for line in self.lines)

    @property
    def start_end(self) -> tuple[int, int]:
        return self.line_start, self.line_end

    @property
    def deprecated(self) -> str:
        deprecation_attributes = ["Deprecated", "DeprecationMessage"]
        for attr in deprecation_attributes:
            prefix = f"{self.indent}\t{attr}:"
            if found := next(
                (line[len(prefix) :].strip() for line in self.lines if line.startswith(prefix)),
                "",
            ):
                return found
        return ""

    def as_dict(self) -> dict[TFSchemaTableColumn, str]:
        return {
            TFSchemaTableColumn.Computability: self.computability,
            TFSchemaTableColumn.Type: self.type,
            TFSchemaTableColumn.Default: self.default,
            TFSchemaTableColumn.PlanModifiers: ", ".join(self.plan_modifiers),
            TFSchemaTableColumn.Deprecated: self.deprecated,
        }

    def row(self, columns: list[TFSchemaTableColumn]) -> list[str]:
        d = self.as_dict()
        return [d[col] for col in columns]

    def parent_attribute_names(self) -> list[str]:
        parent_path = self.attribute_path.removesuffix(self.name).strip(".")
        parents: list[str] = []
        for parent in parent_path.split("."):
            if not parent:
                continue
            if parent.startswith("(") and parent.endswith(")"):
                parents.extend(parent.strip("()").split("|"))
            else:
                parents.append(parent)

        return parents

    def explode(self) -> list[TFSchemaAttribute]:
        exploded = []
        if "|" not in self.absolute_attribute_path:
            return [self]
        part_before, part_after = self.absolute_attribute_path.split("(", maxsplit=1)
        pipe_separated_values, rest = part_after.split(")", maxsplit=1)
        for value in pipe_separated_values.split("|"):
            new_attr = self.model_copy(update={"absolute_attribute_path": f"{part_before}{value}{rest}"})
            new_attr.absolute_attribute_path = f"{part_before}{value}{rest}"
            exploded.append(new_attr)
        return exploded

    def __lt__(self, other) -> bool:
        if not isinstance(other, TFSchemaAttribute):
            raise TypeError
        return self.absolute_attribute_path < other.absolute_attribute_path
